local UnitPosition, GetPlayerFacing, GetPOITextureCoords, GetCVar = UnitPosition, GetPlayerFacing, GetPOITextureCoords, GetCVar
local pi, cos, sin, abs, tinsert, print, type, pairs = math.pi, math.cos, math.sin, math.abs, table.insert, print, type, pairs
local Minimap = Minimap

local MinimapRadius = {
    Inside = {
        [0] = 150.0,
        [1] = 120.0,
        [2] = 90.0,
        [3] = 60.0,
        [4] = 40.0,
        [5] = 25.0,
    },
    Outside = {
        [0] = 233.333324,
        [1] = 199.999992,
        [2] = 166.66666,
        [3] = 133.333328,
        [4] = 99.999996,
        [5] = 66.666664,
    }
}

MinimapPOIHandler = CreateFrame("Frame")
MinimapPOIHandler.POIIcons = {}
MinimapPOIHandler.POIPoints = {}

MinimapPOIHandler:RegisterEvent("MINIMAP_UPDATE_ZOOM")
MinimapPOIHandler:RegisterEvent("CVAR_UPDATE")
MinimapPOIHandler:SetScript("OnEvent", function(self, event, ...) self[event](self, event, ...) end)

function MinimapPOIHandler:AddPoint(point)
    if not point.x or not point.y or not point.name then
        print("MinimapPOIHandler:AddPoint failed")
        return
    end
    tinsert(self.POIPoints, point)
end

function MinimapPOIHandler:SetFilter(filter)
    if type(filter) ~= "function" then
        print("MinimapPOIHandler:SetFilter failed")
        return
    end
    self.filter = filter
end

function MinimapPOIHandler:Enable()
    self:SetScript("OnUpdate", function(self, elapsed) self["UpdateMinimapIcons"](self, elapsed) end)
end

function MinimapPOIHandler:Disable()
    self:SetScript("OnUpdate", nil)
    for k, v in pairs(self.POIIcons) do
        v:Hide()
    end
end

function MinimapPOIHandler:MINIMAP_UPDATE_ZOOM()
    local curZoom = Minimap:GetZoom()
    if GetCVar("minimapZoom") == GetCVar("minimapInsideZoom") then
        if curZoom < 2 then
            Minimap:SetZoom(curZoom + 1)
        else
            Minimap:SetZoom(curZoom - 1)
        end
    end
    if GetCVar("minimapZoom")+0 == Minimap:GetZoom() then
        self.minimapOutside = true
    else
        self.minimapOutside = false
    end
    Minimap:SetZoom(curZoom)
end

function MinimapPOIHandler:CVAR_UPDATE()
    self.rotateMinimap = GetCVar("rotateMinimap") ~= "0"
end

local function MinimapPOI_OnEnter(self)
    GameTooltip:SetOwner(self, "ANCHOR_RIGHT", 25, 0)
    local name = MinimapPOIHandler.POIPoints[self.pointid].name
    GameTooltip:SetText(name, 1, 1, 1, true)
    GameTooltip:Show()
end

local function MinimapPOI_OnLeave(self)
    GameTooltip:Hide()
end

function MinimapPOIHandler:CreatePOI(index, icon)
    local name = "MinimapPOI"..index
    local poi = CreateFrame("Button", name, Minimap)
    poi:SetHeight(20)
    poi:SetWidth(20)

    poi.Texture = poi:CreateTexture(name.."Texture", "BACKGROUND")
    poi.Texture:SetTexture("Interface\\Minimap\\POIIcons")

    local x1, x2, y1, y2 = GetPOITextureCoords(icon)

    poi.Texture:SetTexCoord(x1, x2, y1, y2)

    poi.Texture:SetPoint("CENTER", 0, 0)
    poi.Texture:SetHeight(12)
    poi.Texture:SetWidth(12)

    poi:SetScript("OnEnter", MinimapPOI_OnEnter)
    poi:SetScript("OnLeave", MinimapPOI_OnLeave)

    self.POIIcons[index] = poi

    return poi
end

local function WorldPosToMinimapPos(px, py, x, y, radius)
    local locLeft, locTop, locRight, locBottom = py + radius, px + radius, py - radius, px - radius

    local h, w = abs(locRight - locLeft), abs(locBottom - locTop)
    local sx = (locTop - x) / w
    local sy = (locLeft - y) / h
    local onMap = y < locLeft and y > locRight and x < locTop and x > locBottom
    return sy, sx, onMap
end

function MinimapPOIHandler:GetPOI(index)
    local poi = self.POIIcons[index]
    if not poi then
        poi = self:CreatePOI(index, 197)
    end
    return poi
end

function MinimapPOIHandler:UpdateMinimapIcons(elapsed)
    local x, y, _, mapID = UnitPosition("player")
    local zoom = Minimap:GetZoom()
    local radius = self.minimapOutside and MinimapRadius.Outside[zoom] or MinimapRadius.Inside[zoom]
    local mmw = Minimap:GetWidth()
    local mmh = Minimap:GetHeight()
    local diameter = radius * 2
    local xscale = mmw / diameter
    local yscale = mmh / diameter
    local rotation = (2 * pi) - GetPlayerFacing()

    local index = 0

    for k, v in pairs(self.POIPoints) do
        if self.filter(v, x, y, mapID, radius) then
            index = index + 1
            local poi = self:GetPOI(index)

            poi.pointid = k

            if self.rotateMinimap then
                local dy, dx = v.x - x, v.y - y
                local sinTheta = sin(rotation)
                local cosTheta = cos(rotation)
                local mx = ((-dx * cosTheta) - (dy * sinTheta)) * xscale
                local my = ((-dx * sinTheta) + (dy * cosTheta)) * yscale
                poi:SetPoint("CENTER", Minimap, "CENTER", mx, my)
            else
                local mx, my = WorldPosToMinimapPos(x, y, v.x, v.y, radius)
                mx = mx * mmw
                my = -my * mmh
                poi:SetPoint("CENTER", Minimap, "TOPLEFT", mx, my)
            end

            poi:Show()
        end
    end

    if #self.POIIcons > index then
        for pindex = index + 1, #self.POIIcons do
            self.POIIcons[pindex]:Hide()
        end
    end
end
